#include "hyp2f1.h"
#pragma fp_contract (off)

// pochhammer(z,n) = `z*(z+1)*...*(z+n-1)` = GAMMA(n+z)/GAMMA(z)
// this is a brute version
cplx pochhammer(cplx z, int n)
{
  int k=0;
  cplx p;

  if ( z == ZERO && n == 0 )
    return ONE;
  if ( z == ZERO )  // n <> 0
    return ZERO;
  if ( n < 0 )      // write as Gamma, to see it is true
    return ONE/pochhammer(z + (cplx)n, -n);
  if ( n == 0 )
    return ONE;
  if ( n == 1 )
    return z;
  if ( n == 2 )
    return z * (z + ONE);

  p = z;
  for(k=1; k <= n-1; k++)
    p = p * (z + (cplx)k);

  return (cplx)p;
}

// = (1/(z+1) + GAMMA(z)) - 1/(z+1) in z = -1
// series around -1 up to | z - (-1) | <= 1/128
cplx G_minus1(cplx z)
{
  cplx w = z + ONE;   // close to 0
  cplx w2 = w * w;
  cplx w4 = w2 * w2;
  cplx s = ZERO;  

  // sum up starting with the smallest term 
  // do not use a Horner form in 0 because of very small w
  s =  -0.501241626455653050 * w2*w4;
  s += -1.49724338689808458  * w*w4;
  s += -0.504094272276808391 * w4;
  s += -1.48608934117995359  * w*w2;
  s += -0.504361254345553406 * w2;
  s += -1.41184033042643969  * w;
  s += -0.422784335098467139;
  
  s = s - ONE/w;
  return s;
}

// = (Gamma(z) - 1/z) + 1/z
cplx G_minus0(cplx z)
{
  cplx z2 = z * z;
  cplx z4 = z2 * z2;
  cplx s = ZERO;  

  s =  -0.996001760442431534  * z2*z4 ;
  s += +0.993149114621276193  * z*z4;
  s += -0.981995068903145202  * z4;
  s += +0.981728086834400187  * z*z2;
  s += -0.907479076080886289  * z2;
  s += +0.989055995327972555  * z;
  s += -0.577215664901532861;
  
  s = s + ONE/z;
  return s;
}

// following Godfrey, see his Matlab implementation gamma.m
cplx GAMMA(cplx Z)
{
  cplx z;
  double x;
  cplx zh, zgh, zp;
  cplx sum;
  int k=0;
  cplx result;
  int nx = 0; // integer part of real(z)
  double xi=0.0;

  static double g = 607.0/128.0;
  static double c[] = {
      0.99999999999999709182,
     57.156235665862923517,
    -59.597960355475491248,
     14.136097974741747174,
     -0.49191381609762019978,
      0.33994649984811888699e-4,
      0.46523628927048575665e-4,
     -0.98374475304879564677e-4,
      0.15808870322491248884e-3,
     -0.21026444172410488319e-3,
      0.21743961811521264320e-3,
     -0.16431810653676389022e-3,
      0.84418223983852743293e-4,
     -0.26190838401581408670e-4,
      0.36899182659531622704e-5};
  int dimc = 15; // length of c

  // exact values for GAMMA(n), n = 0 ... 19
  static double G[] ={
    1.0000000000000000e+308, // to indicate overflow
                     1.0,
                     1.0,
                     2.0,
                     6.0,
                    24.0,
                   120.0,
                   720.0,
                  5040.0,
                 40320.0,
                362880.0,
               3628800.0,
              39916800.0,
             479001600.0,
    6.2270208000000000e+09,
    8.7178291200000000e+10,
    1.3076743680000000e+12,
    2.0922789888000000e+13,
    3.5568742809600000e+14,
    6.4023737057280000e+15,
    1.2164510040883200e+17 };

  z = (cplx)Z;
  x = RE(z);
  nx = roundToInteger( RE(z) );
  xi = x - nx;
  

  if ( ( IM(z)== 0.0) && (x == int(x)) )
  {
    if (nx <= 0) return (cplx(1e308, 0.0));   // indicating pole for z = 
    if (nx <= 19) return (cplx)G[nx];         // return exact value, if tabled
  }

  if ( nx < -switchGAMMA )
    return rSIN(z) / GAMMA(-z + ONE);         // reflection for small inputs

  if ( nx == 0 )
  {
    result = (cplx)ONE/pochhammer(z, 1);
    result = result * GAMMA(z + ONE);
    return result;
  }

  if ( nx == -1 && ABS(z + ONE) <= 1.0/128.0 ) // small circle around -1
  {
    result = G_minus1(z);
    return result;
  }

/* not needed
  if ( nx == -2 && ABS(z + ONE+ONE) <= 2.0/128.0 ) // small circle around -2
  {
    result = G_minus2(z);
    return result;
  }
*/
  if ( nx <= -1 && ABS(z - (cplx)nx ) <= 1.0/128.0 )
  {
    result = (cplx)ONE/pochhammer(z, -nx - 1);
    result = result * GAMMA(z - (cplx)nx - ONE);
    return result;
  }
  
  if ( nx <= -1 )
  {
    result = (cplx)ONE/pochhammer(z, -nx + 1);
    result = result * GAMMA(z - (cplx)nx + ONE);
    return result;
  }

  // Lanczos method for Gamma, according to Godfrey
  z = z - ONE;
  
  // his trick for avoiding FP overflow above z=141
  zh = z + 0.5*ONE;
  zgh=zh + (cplx)g;
  zp=POW(zgh, (zh*0.5));
  
  // starting with maximal index and descending, so with increasing summands
  sum = ZERO;
  for (k=dimc - 1; 0<k; k--) 
  {
    sum = sum + (cplx)c[k]/(z + (cplx)k);
  }
  
  result = sqrt2Pi * (sum + (cplx)c[0]); // final, minimal index
  result = result * zp*zp * EXP(-zgh);
  return result;
}
