#include "hyp2f1.h"

// = regularized hypergeometric function = 2F1(a,b,c,z) / Gamma(c)
cplx hyp2F1_regularized (cplx a, cplx b, cplx c, cplx Z)
{
  cplx z=ZERO;
  double x = RE(Z),y = IM(Z);
  cplx result;

  // branch cut
  if ( (1.0 < x) && (fabs(y) <= 1e-306) )
    y = - (1e-306);

  z = x + y*theI;

  result = G_2F1_regularized(a,b,c,z); // 2F1(a,b,c,z)/Gamma(c)

  // clean up branch cut
  if ( fabs(IM(result)) <= 1e-306)
    result = RE(result) + ZERO;

  // the disk, real parameters give real results, if Gamma(c) is real
  if ( (x < 1.0) && (0.0 == y) 
    &&(IM(a) == 0.0) && (IM(b) == 0.0) && (IM(c) == 0.0) && (0.0 <= RE(c)) )
    result = RE(result) + ZERO;

  return result;
}

/*
  value for c = -m negative integer (A&S 15.1.2):
  Limit( F(a,b , c , z) / GAMMA(c), c = -m) = 
  pochhammer(a,m+1)*pochhammer(b,m+1)/(m+1)! * z^(m+1) * F(a+m+1,b+m+1, m+2 ,z)
*/
cplx G_2F1_regularized_exceptional (cplx a, cplx b, int m, cplx z)
{
  cplx result=ZERO;
  cplx M1=ZERO, M2=ZERO;

  M1 = cplx(m+1, 0.0);
  M2 = cplx(m+2, 0.0);
  result = pochhammer(a,m+1)*pochhammer(b,m+1) / GAMMA(M2);
  result = result * POW(z, M1);
  result = result * G_2F1(a+M1,b+M1 , M2 , z);
  return result;
}

/*
  Take a circle, centered in that problematic parameter of small radius and
  choose equal distant points on the circle line. Then F(c)/Gamma(c) is the
  average in that points. This follows from Cauchy's formula (since that fct
  is the regularized 2F1 and analytic), if one writes down the curve integral
  and applies Simpson's rule for integration. 
  
  This generally is not very exact for arbitrary points in the circle - but it
  is fine for the center, so a few points are sufficient.

  If we *assume* that Numerics up to 0.001 <= |c - pole| is good enough for an
  existing code, then one can try 1/256 ~ 0.004 for the radius: all points on
  the circle line will have a distance beyond 0.003.

  Note that it is no longer that fine for the derivative in general.
*/

// not used for G_2F1
cplx G_2F1_regularized_close_exceptional (cplx a, cplx b, cplx c, cplx z)
{
  cplx center=c;
  cplx s=ZERO;
  int k=0;

  center = c;
  for(k=0;k<degROOTS/2;k++) // the other roots are the negative ones
    s = s + 
      G_2F1(a,b, +ROOTS[k]*1.0 + center, z) * rGAMMA(+ROOTS[k]*1.0 + center) + 
      G_2F1(a,b, -ROOTS[k]*1.0 + center, z) * rGAMMA(-ROOTS[k]*1.0 + center);
  
  s = 1.0/(double)degROOTS *s; // average around center
  return s;
}

cplx G_2F1_regularized (cplx a, cplx b, cplx c, cplx z)
{
  //return G_2F1(a,b,c, z) * rGAMMA(c);
  cplx C=ZERO;
  int m=0;
  cplx result=ZERO;
  cplx M1=ZERO, M2=ZERO;

  /* value for c = -m negative integer (A&S 15.1.2):
     Limit( F(a,b , c , z) / GAMMA(c), c = -m) = 
     pochhammer(a,m+1)*pochhammer(b,m+1)/(m+1)! * z^(m+1) * F(a+m+1,b+m+1, m+2, z)
  */
  
  //if ( RE(c) < 0.5 && -0.5 < IM(c) && IM(c) < 0.5 )
  m = - roundToInteger(RE(c)); // m = 0,1,2, ...

  if ( -RE(c) == m && IM(c) == 0.0 ) // i.e. c = -m + 0*I
    return G_2F1_regularized_exceptional(a,b,-m,z);

  if ( ABS(c + cplx((double)m,0.0)) < regRADIUS )
    return G_2F1_regularized_close_exceptional(a,b,c,z);

  return G_2F1(a,b,c, z) * rGAMMA(c);
}
