#include "hyp2f1.h"
#pragma fp_contract (off)

// Euler's transform, no restrictions for parameters or variable
cplx Euler (cplx a, cplx b, cplx c, cplx z)     // = A&S 15.3.3
{
  return POW(ONE - z, c-a-b) * (cplx)G_2F1(c - b, c - a, c, z);
}

// polynomial case only to be used, if c-b or c-a are integers <= 0
cplx Euler_polynom (cplx a, cplx b, cplx c, cplx z)
{
  //return POW(ONE - z, c - a - b) * 
    //(cplx)polynom_2F1(c - b, c - a, c, z, 1111);
  return POW(ONE - z, c - a - b) * 
    (cplx)polynomialseries_2F1(c - b, c - a, c, z);
}

// Pfaff's transform, (1-z)^(-a)*hypergeom([a, c-b],[c],z/(z-1)),  w = z/(z-1)
// no restrictions for parameters, but z off from 1, A&S 15.3.5
cplx Pfaff (cplx a, cplx b, cplx c, cplx z)
{
  cplx result;
  if ( ABS(a*(c - b)) < ABS((c - a)*b) )            // use A&S 15.3.4
    result = POW(ONE - z, - a) * (cplx)G_2F1(a, c - b, c, z/(z-ONE));
  else                                              // use A&S 15.3.5
    result = POW(ONE - z, - b) * (cplx)G_2F1(c - a, b, c, z/(z-ONE));
  return(result);
}

//############################################################################

/*
  The main linear transforms (not as in Abramowitz & Stegun, but as in the
  proofs of Lebedev's book "Special Functions and their Applications") for
  the case of non-exceptional parameters:

  Transformed = GAMMA(c) * PI/sin(PI*q) * g(a,b,c,z) where g is always a
  difference (as in A&S), q = linear combination of the parameters.
  
  Note that this can lead to so called  "catastrophic cancellations" errors
  (same magnitude, opposite sign, hence only few final decimal places survive,
  which is not enough for precise results or even quite false in sequel).
  That is independent from prior exact results and just depending on input.

  The most reasonably way is to use a few as possible transform, having just
  a single one is desirably.

  Also observe, that for c ~ negative integer this may become almost useless
  by the poles of GAMMA, especially if c becomes smaller
*/

// this is in Lebedev, p.256, but with slightly different notations
cplx g_15_3_6 (cplx a, cplx b, cplx c, cplx z)
{
  cplx w, mu, f1, f2, g1, g2;

  w =  (ONE - z);
  mu = (c - a - b);
  f1 = rGAMMA(b + mu) * rGAMMA(a + mu);
  f2 = POW(w, mu) * rGAMMA(a) * rGAMMA(b);
  //g1 = G_2F1(a, b, -mu + ONE, w) * rGAMMA(-mu + ONE) ;
  //g2 = G_2F1(a + mu, b + mu, ONE + mu, w) * rGAMMA(ONE + mu);
  g1 = G_2F1_regularized(a, b, -mu + ONE, w);
  g2 = G_2F1_regularized(a + mu, b + mu, ONE + mu, w);
  return (cplx)(f1 * g1 - f2 * g2);
}

cplx g_15_3_7 (cplx a, cplx b, cplx c, cplx z)
{
  cplx w, mu, f1, f2, g1, g2;

  w =  (ONE / z);
  mu = (b - a);
  f1 = POW(-z, -a) * rGAMMA(c - a) * rGAMMA(b);
  f2 = POW(-z, -b) * rGAMMA(a) * rGAMMA(c - b);
  //g1 = G_2F1(a, ONE + a - c, -mu + ONE, w) * rGAMMA(-mu + ONE);
  //g2 = G_2F1(b, b - c + ONE, mu + ONE, w) * rGAMMA(mu + ONE);
  g1 = G_2F1_regularized(a, ONE + a - c, -mu + ONE, w);
  g2 = G_2F1_regularized(b, b - c + ONE, mu + ONE, w);
  return (cplx)( f1 * g1 - f2 *g2 );
}

cplx g_15_3_9 (cplx a, cplx b, cplx c, cplx z)
{
  cplx w, mu, f1, f2, g1, g2;

  w =  (z - ONE);
  w =  (w / z);    // w:= (z-1)/z;
  mu = (c - a - b);
  f1 = POW(z, -a) * rGAMMA(b + mu) * rGAMMA(a + mu);
  f2 = POW(ONE - z, mu) * POW(z, -b - mu) * rGAMMA(b) * rGAMMA(a);
  //g1 = G_2F1(a, ONE - b - mu, -mu + ONE, w) * rGAMMA(-mu + ONE);
  //g2 = G_2F1(b + mu, ONE - a, mu + ONE, w) * rGAMMA(mu + ONE);
  g1 = G_2F1_regularized(a, ONE - b - mu, -mu + ONE, w);
  g2 = G_2F1_regularized(b + mu, ONE - a, mu + ONE, w);
  return (cplx)( f1 * g1 - f2 *g2 );
}

//############################################################################

/*
  Exceptional parameters: numerical differentiation w.r.t. parameters avoiding
  the usual series in 4 Psi functions given in Abramowitz & Stegun.

  Stepsize is fixed and quite large to keep off singularities (but can be set
  through the header file if needed), a Lagrange version of degree 8 is used.
  
  This is done along the complex direction, so the line through the points will
  avoid the poles and thus F is analytic around the path.
*/

static cplx Lagrange8[] = {
    +1.0/280.0,
    -4.0/105.0,
    +1.0/5.0,
    -4.0/5.0,
    +4.0/5.0,
    -1.0/5.0,
    +4.0/105.0,
    -1.0/280.0
  };

cplx exceptional_15_3_6 (
  cplx a,
  cplx b,
  cplx C, // C
  cplx z)
{
  int m;
  double mu;
  cplx c;
  cplx result;
  cplx h;
  cplx dg;
  cplx x;

  mu = (RE(C - a - b));
  m = roundToInteger(mu);
  c = cplx((double)m,0.0) + a + b;
  x = c;

  h = STEPSIZE * theI;
  dg =
    + Lagrange8[0]  * (g_15_3_6(a,b,x-4.0*h,z) - g_15_3_6(a,b,x+4.0*h,z))
    + Lagrange8[1]  * (g_15_3_6(a,b,x-3.0*h,z) - g_15_3_6(a,b,x+3.0*h,z))
    + Lagrange8[2]  * (g_15_3_6(a,b,x-2.0*h,z) - g_15_3_6(a,b,x+2.0*h,z))
    + Lagrange8[3]  * (g_15_3_6(a,b,x-1.0*h,z) - g_15_3_6(a,b,x+1.0*h,z));

  dg = dg / h;

  result = GAMMA(c) * dg;
  
  if (isOdd(m) == 1)
    result = -result;
  return(result);
}

cplx exceptional_15_3_7 (
  cplx a,
  cplx B, // B
  cplx c,
  cplx z)
{
  int m;
  double mu;
  cplx b;
  cplx result;
  cplx h;
  cplx dg;
  cplx x;

  mu =(RE(B - a));
  m = roundToInteger(mu);
  b = cplx((double)m,0.0) + a;
  x = b;

  h = STEPSIZE * theI;
  dg =
    + Lagrange8[0]  * (g_15_3_7(a,x-4.0*h,c,z) - g_15_3_7(a,x+4.0*h,c,z))
    + Lagrange8[1]  * (g_15_3_7(a,x-3.0*h,c,z) - g_15_3_7(a,x+3.0*h,c,z))
    + Lagrange8[2]  * (g_15_3_7(a,x-2.0*h,c,z) - g_15_3_7(a,x+2.0*h,c,z))
    + Lagrange8[3]  * (g_15_3_7(a,x-1.0*h,c,z) - g_15_3_7(a,x+1.0*h,c,z));

  dg = dg / h;

  result = GAMMA(c) * dg;
  
  if (isOdd(m) == 1)
    result = -result;
  return(result);
}

cplx exceptional_15_3_9 (
  cplx a,
  cplx b,
  cplx C, // C
  cplx z)
{
  int m;
  double mu;
  cplx c;
  cplx result;
  cplx h;
  cplx dg;
  cplx x;

  mu = RE(C - a - b);
  m = roundToInteger(mu);
  c = cplx((double)m, 0.0) + a + b;
  x = c;

  h = STEPSIZE * theI;
  dg =
    + Lagrange8[0]  * (g_15_3_9(a,b,x-4.0*h,z) - g_15_3_9(a,b,x+4.0*h,z))
    + Lagrange8[1]  * (g_15_3_9(a,b,x-3.0*h,z) - g_15_3_9(a,b,x+3.0*h,z))
    + Lagrange8[2]  * (g_15_3_9(a,b,x-2.0*h,z) - g_15_3_9(a,b,x+2.0*h,z))
    + Lagrange8[3]  * (g_15_3_9(a,b,x-1.0*h,z) - g_15_3_9(a,b,x+1.0*h,z));

  dg = dg / h;

  result = GAMMA(c) * dg;
  
  if (isOdd(m) == 1)
    result = -result;
  return(result);
}

//############################################################################

/*
  Bring it together:

  For exceptional constellations of parameters use from differentiation above.

  If the parameter constellation is only very close to the exceptional case,
  then one can average around small circles around the 1st parameter, see the
  comments for the regularized 2F1.

  Beyond that one can use linear transform formulae.

  Note however, that these substract 2 terms and "catastrophic cancellation"
  may occure: if the two terms of almost the same value, the 'heads' vanish
  and the   'tails' remain - but they are too short to represent the actual
  value.

  One may try to heal it by using the series form for them (the transforms
  are used such that it would be mostly possible). But experiments show that
  there are constellations where this shit already happens for the lowest
  summands. Thus I did not go that way, may be others do that work.
*/

cplx AS_15_3_6_Leb (cplx a, cplx b, cplx c, cplx z)   // w = 1 - z
{
  cplx mu = c - a - b;
  double m = (double)roundToInteger(RE(mu));
  cplx center=a;
  cplx s=ZERO;
  int k=0;

  if ( IM(mu) == 0.0 && RE(mu) == m ) // mu = m + 0*I
    return exceptional_15_3_6(a, b, c, z);

  // close to exceptional
  if ( ABS(mu - cplx(m,0.0) ) < regRADIUS ) // if ( ABS(c-a-b - m ) < regRADIUS )
  {
    for(k=0;k<degROOTS/2;k++) // the other roots are the negative ones
      s = s + 
        AS_15_3_6_Leb(+ROOTS[k] + center,b,c, z) + 
        AS_15_3_6_Leb(-ROOTS[k] + center,b,c, z);
  
    s = 1.0/(double)degROOTS *s; // average around center
    return s;
  }
  
  return(rSIN(mu) * GAMMA(c) * g_15_3_6(a,b,c,z));
}

cplx AS_15_3_7_Leb (cplx a, cplx b, cplx c, cplx z)   // w = 1/z
{
  cplx mu = b-a;
  double m = (double)roundToInteger(RE(mu));
  cplx center=a;
  cplx s=ZERO;
  int k=0;

  if ( IM(mu) == 0.0 && RE(mu) == m ) // mu = m + 0*I
    return(exceptional_15_3_7(a, b, c, z));
  
  //if ( ABS(a-b - m ) < regRADIUS )
  if ( ABS(mu - cplx(m, 0.0) ) < regRADIUS )
  {
    for(k=0;k<degROOTS/2;k++) // the other roots are the negative ones
      s = s + 
        AS_15_3_7_Leb(+ROOTS[k] + center,b,c, z) + 
        AS_15_3_7_Leb(-ROOTS[k] + center,b,c, z);
  
    s = 1.0/(double)degROOTS *s; // average around center
    return s;
  }

  return(rSIN(mu) * GAMMA(c) * g_15_3_7(a,b,c,z));
}

cplx AS_15_3_9_Leb (cplx a, cplx b, cplx c, cplx z)   // w = 1 - 1/z
{
  cplx mu = c - a - b;
  double m = (double)roundToInteger(RE(mu));
  cplx center=a;
  cplx s=ZERO;
  int k=0;

  if ( IM(mu) == 0.0 && RE(mu) == m ) // mu = m + 0*I
    return(exceptional_15_3_9(a, b, c, z));

  if ( ABS(mu - cplx(m, 0.0) ) < regRADIUS ) //if ( ABS(c-a-b - m ) < regRADIUS )
  {
    for(k=0;k<degROOTS/2;k++) // the other roots are the negative ones
      s = s + 
        AS_15_3_9_Leb(+ROOTS[k] + center,b,c, z) + 
        AS_15_3_9_Leb(-ROOTS[k] + center,b,c, z);
  
    s = 1.0/(double)degROOTS *s; // average around center
    return s;
  }
  
  return(rSIN(mu) * GAMMA(c) * g_15_3_9(a,b,c,z));
}

