#include "hyp2f1.h"
#pragma fp_contract (off)

// round to nearest integer
int roundToInteger (double x)
{
  if (0.0 < x)
    return (int)floor(x + 0.5);
  return (int)ceil(x - 0.5);
}

int isEven (int m)
{
  if (abs(m) % 2 == 0)
  // otherwise modulo may have negative sign in Microsoft's Numerics
  // and for mod2 this is enough
    return (1);
  else
    return (0);
}

int isOdd (int m)
{
  if (abs(m) % 2 == 0)
    return (0);
  else
    return (1);
}

// fcmp in GSL: equal w.r.t. relative accuracy (i.e. regarding magnitude)
// implementation following T.C. Belding, based on D.E. Knuth
// return = 0 if 'are equal', else return = 1
int areClose (const double x1, const double x2, const double epsilon)
{
  int exponent;
  double delta, difference, maxXY, dummy;

  /* Find exponent of largest absolute value */
  maxXY = (fabs (x1) > fabs (x2)) ? x1 : x2;
  dummy = frexp (maxXY, &exponent);

  /* Form a neighborhood of size  2 * delta */
  delta = ldexp (epsilon, exponent);
  difference = x1 - x2;

  if (difference > delta)       /* x1 > x2 */
      return 0;
  else if (difference < -delta) /* x1 < x2 */
      return 0;
  else                          /* -delta <= difference <= delta */
      return 1;                 /* x1 ~=~ x2 */
}

int areEqual (cplx a, cplx b)
{
  if ( real(a) == real(b) && 
        imag(a) == imag(b) )
    return(1);

  // otherwise care for magnitudes
  if ( areClose(real(a), real(b), DBL_MIN) && 
        areClose(imag(a), imag(b), DBL_MIN) )
    return(1);
  
  return(0);
}

// z = 0, +-1, +-2, ... ---> return 1
int isNumericalInteger (cplx z)
{
  double x = fabs(RE(z));

  if ( 0.0 == IM(z) )
  {
    if ( roundToInteger(x) == x )
      return(1);
  }

  return(0);
}

// z = 0, 1, 2, ... ---> return 1, for me Naturals start in 0
int isNumericalNaturalNumber (cplx z)
{
  if ( (IM(z) == 0.0 && 0.0 <= RE(z)) && (isNumericalInteger(z) == 1) )
    return(1);
  return(0);
}


double maxNORM (cplx z) // = maximum norm
{
  double x = fabs(real(z));
  double y = fabs(imag(z));
  return((x >= y ? x : y));
}

// ############################################################################################

static double DP1 = 3.14159265160560607910E0;
static double DP2 = 1.98418714791870343106E-9;
static double DP3 = 1.14423774522196636802E-17;

/*
  Cody & Waite 
  returns x modulo Pi, ie: x = xi + k*PI, 0 <= |xi| < PI and k integer
  = redupi of Moshier's Cephes 
  or GSL for Pi/2 (gsl_sf_sin_e) or also SUN's fdlibm
  There are certainly better range reductions (Payne&Hanek? Muller?)
  but it is exact except the last place for |x| <= 10^5, which is more
  then what is needed here and it is quite handy.
*/
double moduloPI(double x) 
{
  double t;
  long k;

  if ( fabs(x) < PI) return x;

  t = x/PI;
  if( t >= 0.0 )
    t += 0.5;
  else
    t -= 0.5;

  k = (long)t;	/* the multiple */
  t = (double)k;
  t = ((x - t * DP1) - t * DP2) - t * DP3;
  // ensure that order of evaluation, the compiler should NOT optimize
  // that away to give t = x - k*(DP1+DP2+DP3) = x - k*Pi
  return (t);
}

// there *should* be no need to use this range reduction
cplx rSIN (cplx z)  // = PI / sin( PI * z)
{
  cplx W_red=ZERO;
  double X=0.0, X_red=0.0;
  double x=0.0, y=0.0, t=0.0;
  int k=0;
  cplx result=ZERO;

  x = RE(z); y = IM(z);

  if ( x == 0.0 ) // sin(y*I) = sinh(y)*I, Pi/sin(Pi*y*I) = -I*Pi/sinh(Pi*y)
  {
    t = PI * y;
    result = - PI / sinh(PI * y);
    result = result * theI;
    return (result); // purely imaginary
  }

  // range reduction
  X = x*PI;
  X_red = moduloPI(X); // X = X_red + k*Pi
  k = roundToInteger( (X - X_red)/PI );
  
  W_red = X_red + y*PI *theI;

  result = PI / SIN(W_red);
  if ( isOdd(k) == 1 ) //if ( isEven(m) == 1 )
    result = - result;
  
  if (y == 0.0 )  // purely real
    result = RE(result) + 0.0*theI;
  return (result);
}
